#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <stdbool.h>
#include <omp.h>
#include "localmath.h"

#define PI 3.141592653589793238
#define TWOPI 6.28318530718
#define MAXNLAMBDA 20


int compare_t (const void * a, const void * b)
{
  if ( *(double *) a < *(double *) b ) return -1;
  if ( *(double *) a > *(double *) b ) return 1;
  return 0;
}





double calc_exp_time(int ns, int np, int istar, int ilambda, int nphases, int norbits, double IWA, double OWA, double min_deltamag, \
		     double *Istar, double *noisefloor, double *photap_frac, double *omega_lod, double *skytrans, int npix, double xcenter, double ycenter, double pixscale, double bw, double *angdiams, int ndiams, int npsfratios, int nrolls, double *psf_trunc_ratio,  \
		     double lambda, double D, double A, double magstar, double F0, double throughput, double Fzodi, double Fexozodi, double Fbinary, double R, double SNR, \
		     double det_pixscale_mas, double det_npix_multiplier, double det_DC, double det_RN, double det_tread, double det_CIC, \
		     double dist, double angdiamstar_arcsec, double toverhead_fixed, double toverhead_multi, double tlimit, int optimize_phase, double *Fp0, double *sp, double *xp, double *yp, double *tp, \
		     double *avgpsfomega, double *minpsfomega, double *maxpsfomega, double *avgpsftruncratio, double *minpsftruncratio, double *maxpsftruncratio, int printcountrates) {



  //clock_t t0, t1, t2, t3, t4, t5, t6, t7, t8, t9;
  //t0 = clock();

  
  int iplanet, iplanetpistartnp, istartnp, iorbit, iphase, iratio, bestiratio;
  long istartnpix2, index, index2;
  double Fstar;
  double CRp, CRb;
  double CRpfactor, CRbzfactor, CRbezfactor, CRbsfactor, CRbbinfactor, CRbdfactor;
  double tempCRpfactor, tempCRbsfactor, tempCRbzfactor, tempCRbezfactor, tempCRbbinfactor, tempCRbfactor, tempCRnffactor;
  int CRbflag;
  double CRbs, CRbz, CRbez, CRbbin, CRbd, CRnoisefloor, twoCRb, cp;
  double det_CRp, det_CRbs, det_CRbz, det_CRbez, det_CRbbin, det_CR;
  double det_photap_frac, det_omega_lod, det_Istar, det_skytrans, det_sep, det_sep_pix;
  double detpixscale_lod, detpixscale_lod2, oneodetpixscale_lod2, det_npix, tempdetnpix;
  double ix, iy;
  double INF = 1.0/0.0;
  double deltalambda_nm; //bandpass for exposure time calculation in microns
  double FATDL;
  double sp_lod;
  int i, j, k;
  double s, dx, dy;
  double maxphotap_frac;
  double stellar_diam_lod;
  int npix2;
  double cot, maxcot, mint_v_phase, minval;
  double SNR2;
  double SNRCRpfloor;
  double lod_arcsec, lod_arcsec2, oneolod_arcsec;
  double oneopixscale2, pixscale_rad, oneopixscale_arcsec;
  double besttp_v_ratio, bestpsfomega, best_psfomega_v_phase, bestpsftruncratio, best_psftruncratio_v_phase;
  double temptp;
  double A_cm;
  double t_photon_count;
  int psfcounter;
  

  //Define some quantities up front
  cot = 0.0;
  maxcot = 0.0;
  npix2 = npix * npix;
  istartnp = istar*np;
  istartnpix2 = istar * npix * npix;
  SNR2 = SNR*SNR;
  A_cm = A * 100. * 100.; //collecting area of the telescope (cm^2)
  lod_arcsec = (lambda*1e-6 / D) * 206264.806; //l/D in arcsec
  oneolod_arcsec = 1.0 / lod_arcsec;
  lod_arcsec2 = lod_arcsec * lod_arcsec;
  pixscale_rad = pixscale * (lambda*1e-6 / D); //pixel scale in radians    
  oneopixscale_arcsec = 1.0 / (pixscale_rad * 206264.806); //one over pix scale in arcsec
  oneopixscale2 = 1.0 / (pixscale * pixscale); //(l/D)^2
  detpixscale_lod = det_pixscale_mas / (lod_arcsec*1000.);
  detpixscale_lod2 = detpixscale_lod * detpixscale_lod;
  oneodetpixscale_lod2 = 1.0 / detpixscale_lod2;

  
  
  //Define some arrays that we use locally for interpolation
  double *Istar_interp;
  Istar_interp = (double *) malloc (npix2 * sizeof(double));
  if (Istar_interp==NULL) {printf(stderr,"ERROR: memory acquisition failed when creating Istar_interp array.\n"); exit (1);}
  double *noisefloor_interp;
  noisefloor_interp = (double *) malloc (npix2 * sizeof(double));
  if (noisefloor_interp==NULL) {printf(stderr,"ERROR: memory acquisition failed when creating noisefloor_interp array.\n"); exit (1);}
  double *temp_tp;
  temp_tp = (double *) malloc (np * sizeof(double));
  if (temp_tp==NULL) {printf(stderr,"ERROR: memory acquisition failed when creating temp_tp array.\n"); exit (1);}
  double *psfomegas;
  psfomegas = (double *) malloc (np * sizeof(double));
  if (psfomegas==NULL) {printf(stderr,"ERROR: memory acquisition failed when creating psfomegas array.\n"); exit (1);}
  for(iplanet=0;iplanet<np;iplanet++) psfomegas[iplanet] = 0.0;
  double *psftruncratios;
  psftruncratios = (double *) malloc (np * sizeof(double));
  if (psftruncratios==NULL) {printf(stderr,"ERROR: memory acquisition failed when creating psftruncratios array.\n"); exit (1);}
  for(iplanet=0;iplanet<np;iplanet++) psftruncratios[iplanet] = 0.0;

  //t1 = clock();

  
  //INTERPOLATE ISTAR AND NOISEFLOOR TO GIVEN STELLAR ANGULAR DIAMETER
  //This must be done for each star, each coronagraph, and each wavelength
  stellar_diam_lod = angdiamstar_arcsec * oneolod_arcsec; 
  for(k=0;k<ndiams;k++) {
    if(stellar_diam_lod < angdiams[k]) break;
  }
  
  for(i=0;i<npix;i++) {
    for(j=0;j<npix;j++) {
      //Istar interpolation
      if(k < ndiams) {
	//linear interpolate between (k-1)st and kth angdiam
	Istar_interp[i+j*npix] = Istar[i+j*npix+(k-1)*npix2] + (Istar[i+j*npix+k*npix2] - Istar[i+j*npix+(k-1)*npix2]) * ((stellar_diam_lod-angdiams[k-1])/(angdiams[k]-angdiams[k-1]));
      } else {
	//linear extrapolation beyond the provided ang diams
	Istar_interp[i+j*npix] = Istar[i+j*npix+(k-1)*npix2] + (Istar[i+j*npix+(k-1)*npix2] - Istar[i+j*npix+(k-2)*npix2]) * ((stellar_diam_lod-angdiams[k-1])/(angdiams[k-1]-angdiams[k-2]));
      }
      if(Istar_interp[i+j*npix] < 0) Istar_interp[i+j*npix] = 0.0; //ensure it's >= 0
      
      //Noise floor interpolation
      if(k < ndiams) {
	//linear interpolate between (k-1)st and kth angdiam
	noisefloor_interp[i+j*npix] = noisefloor[i+j*npix+(k-1)*npix2] + (noisefloor[i+j*npix+k*npix2] - noisefloor[i+j*npix+(k-1)*npix2]) * ((stellar_diam_lod-angdiams[k-1])/(angdiams[k]-angdiams[k-1]));
      } else {
	//linear extrapolation beyond the provided ang diams
	noisefloor_interp[i+j*npix] = noisefloor[i+j*npix+(k-1)*npix2] + (noisefloor[i+j*npix+(k-1)*npix2] - noisefloor[i+j*npix+(k-2)*npix2]) * ((stellar_diam_lod-angdiams[k-1])/(angdiams[k-1]-angdiams[k-2]));
      }
      if(noisefloor_interp[i+j*npix] < 0) noisefloor_interp[i+j*npix] = 0.0; //ensure it's >= 0
    }
  }
  //t2 = clock();
  
  //MEASURE CORONAGRAPH PERFORMANCE AT IWA 
  //We will use these to estimate detector count rates if necessary
  //First, find the value of psf_trunc_ratio closest to 0.3
  minval = 1e20;
  bestiratio = 0;
  for(iratio=0;iratio<npsfratios;iratio++) {
    if(fabs(psf_trunc_ratio[iratio]-0.3) < minval) {
      minval = fabs(psf_trunc_ratio[iratio]-0.3);
      bestiratio = iratio;
    }
  }
  iratio = bestiratio;
  //Now find the IWA
  maxphotap_frac = 0.0;
  for(i=0;i<(npix/2);i++) { //Find the maximum throughput
    index2 = i+((int) floor(ycenter))*npix + iratio*npix2; //index for the photap_frac and omega_lod arrays
    if(photap_frac[index2] > maxphotap_frac) {
      maxphotap_frac = photap_frac[index2];
    }
  }
  //Now go back and find where it falls to half it's max value
  //start in the middle of the image
  for(i=((int) floor(xcenter));i>0;i--) { //Find the maximum throughput
    index2 = i+((int) floor(ycenter))*npix + iratio*npix2; //index for the photap_frac and omega_lod arrays
    if(photap_frac[index2] > 0.5*maxphotap_frac) { //we found it
      det_sep_pix = (fabs((i+0.5)-xcenter)); //the separation in pix
      det_sep = det_sep_pix/oneopixscale_arcsec; //record the separation in arcsec
      break;
    }
  }
  //Now calculate the max of Istar at the IWA
  det_Istar = 0.0;
  det_skytrans = 0.0;
  det_photap_frac = 0.0;
  det_omega_lod = 0.0;
  for(i=0;i<npix;i++) {
    for(j=0;j<npix;j++) {
      dx = ((i+0.5)-xcenter);
      dy = ((j+0.5)-ycenter);
      s=sqrt(dx*dx+dy*dy);
      if(fabs(s-det_sep_pix) < 2) { //we have a 2 pixel annulus at det_sep
	index = i+j*npix;
	index2 = index + iratio*npix2; //index for the photap_frac and omega_lod arrays
	if(Istar_interp[index] > det_Istar) det_Istar = Istar_interp[index]; //record the max in the annulus
	if(skytrans[index] > det_skytrans) det_skytrans = skytrans[index]; //record the max skytrans
	if(photap_frac[index2] > det_photap_frac) {
	  det_photap_frac = photap_frac[index2]; //record the max photapfrac
	  det_omega_lod = omega_lod[index2];
	}
      }
    }
  }
  

  
  //t3 = clock();
  
  //CALCULATIONS DONE ONCE AT EACH WAVELENGTH
  //Note: for a 2D array entry c[ix,iy], the 1D index = ix + iy*xsize
  //deltalambda depends on coronagraph and wavelength
  deltalambda_nm = MIN((lambda*1000.0) / R, bw * (lambda*1000.0)); //take the lesser of the desired bandwidth and what coronagraph allows
  FATDL = F0 * A_cm * throughput * deltalambda_nm;
  Fstar = pow((double) 10., -0.4*magstar);
  CRpfactor = Fstar * FATDL; //for planet's count rate calculation
  CRbzfactor = Fzodi * lod_arcsec2 * FATDL;  //count rate for zodi 
  CRbezfactor = Fexozodi * lod_arcsec2 * FATDL / (dist*dist); //count rate for exozodi -- the dist^2 is there to convert sp (in arcsec) to AU later on...
  CRbsfactor = Fstar * oneopixscale2 * FATDL; //for stellar leakage count rate calculation  
  CRbbinfactor = Fbinary * FATDL; //count rate for scattered light from nearby stars 
  
  //Here we calculate detector noise, as it may depend on count rates
  //We don't know the count rates yet, so we make estimates based on
  //values near the IWA
  det_CRp = 10. * pow((double) 10., -0.4*min_deltamag) * CRpfactor * det_photap_frac; //we budget for 10x the planet count rate
  det_CRbs = CRbsfactor * det_Istar; 
  det_CRbz = CRbzfactor * det_skytrans;
  det_CRbez = CRbezfactor * det_skytrans / (det_sep*det_sep);
  det_CRbbin = CRbbinfactor * det_skytrans;
  
  det_CR = det_CRp + det_CRbs + det_CRbz + det_CRbez + det_CRbbin;
  //According to Bernie Rauscher:
  // effective_dark_current = dark_current - f * cic * (1+W_-1[q/e])^-1.
  // If q = 0.99, (1+W_-1[q/e])^-1 = -6.73 such that
  // effective_dark_current = dark_current + f * cic * 6.73,
  // where f is the brightest pixel you care about in counts s^-1
  det_npix = det_npix_multiplier * det_omega_lod * oneodetpixscale_lod2; //this is temporary to estimate the per pixel noise
  t_photon_count = 1./(6.73 * (det_CR/det_npix));
  CRbdfactor = det_DC + det_RN * det_RN/det_tread + det_CIC / t_photon_count; //need to multiply be det_npix later...

  if(printcountrates == 1) printf("%d | Basic params:   %e %e %e %e %e %e\n",istar,F0,magstar,D,A_cm,lambda*1000.,deltalambda_nm);
  if(printcountrates == 1) printf("%d | Basic params:   %e %e %e %e %e %e %e %e %e %e %e\n",istar,SNR,toverhead_fixed,toverhead_multi,det_DC,det_RN,det_CIC,det_tread,det_pixscale_mas,0.75,0.9,throughput/(0.9*0.75));
  

  //t4 = clock();
  
  //----------------------------------
  //EXPOSURE TIME
  for(iorbit=0;iorbit<norbits;iorbit++) {
    mint_v_phase = INF;
    best_psfomega_v_phase = 0.0;
    best_psftruncratio_v_phase = 0.0;
    
    for(iphase=0;iphase<nphases;iphase++) {

      iplanet = iphase + iorbit*nphases;
      CRbflag = 0;
    
      
      //CALCULATIONS DONE FOR EACH SYNTHETIC PLANET
      //count rate for planet
      //for a 3D array entry a[ix,iy,iz], the 1D index = ix + iy*xsize + iz*(xsize*ysize)
      iplanetpistartnp = iplanet + istartnp;
      ix = xp[iplanetpistartnp] * oneopixscale_arcsec + xcenter; 
      iy = yp[iplanetpistartnp] * oneopixscale_arcsec + ycenter;
      sp_lod = sp[iplanetpistartnp] * oneolod_arcsec;


      
      //if planet is within the boundaries of the coronagraph simulation and hard IWA/OWA cutoffs...
      if(ix >= 0 && ix < npix && iy >= 0 && iy < npix && sp_lod > IWA && sp_lod < OWA) {

	
	index = (long) floor(ix + floor(iy)*npix);
	
	tempCRpfactor = Fp0[iplanetpistartnp] * CRpfactor;
	
	//Calculate the count rate of the noise floor.
	//This should be the stddev (over the "noise region") of
	//the difference of the photometric aperture-integrated
	//stellar PSFs. The photometric aperture integration, and
	//stddev of that have been calculated prior to the call to
	//this function, and was then divided by the number of pixels
	//in the photometric aperture. So here we calculate noise
	//using the same method as the leaked starlight.
	//The "noisefloor" array is equal to
	//stddev(integral(Istar1,dphotometric_ap) - integral(Istar2,dphotometric_ap)) / (omega/(npix*npix))
	tempCRnffactor = SNR * CRbsfactor * noisefloor_interp[index];
	
	
	besttp_v_ratio = INF;
	bestpsfomega = 0.0;
	bestpsftruncratio = 0.0;
	for(iratio=0;iratio<npsfratios;iratio++) {
	  
	  index2 = index + iratio * npix2; //index for the photap_frac and omega_lod arrays
	  
	  //First we just calculate CRp and CRnoisefloor to see if CRp > CRnoisefloor
	  //If CRp > CRnoisefloor, then we will calculate the background
	  CRp = tempCRpfactor * photap_frac[index2];      
	  CRnoisefloor = tempCRnffactor * omega_lod[index2];

	  
	  //If it's above the noise floor, calculate the exposure time...
	  SNRCRpfloor = CRnoisefloor;

	  if(CRp > SNRCRpfloor && omega_lod[index2] > detpixscale_lod2) {
	    
	    if(CRbflag != 1) { //background factors not yet calculated
	      tempCRbsfactor = CRbsfactor * Istar_interp[index];
	      tempCRbzfactor = CRbzfactor * skytrans[index];
	      tempCRbezfactor = CRbezfactor * skytrans[index] / (sp[iplanetpistartnp]*sp[iplanetpistartnp]); //correct for 1/r^2 illumination for exozodi
	      tempCRbbinfactor = CRbbinfactor * skytrans[index];
	      tempCRbfactor = (tempCRbsfactor + tempCRbzfactor + tempCRbezfactor + tempCRbbinfactor);
	      CRbflag = 1; //mark them as calculated
	    }
	    //CRbs = tempCRbsfactor * omega_lod[index2]; 
	    //CRbz = tempCRbzfactor * omega_lod[index2]; 
	    //CRbez = tempCRbezfactor * omega_lod[index2];
	    //CRbbin = tempCRbbinfactor * omega_lod[index2];
	    CRb = tempCRbfactor * omega_lod[index2]; 
	    det_npix = det_npix_multiplier * (omega_lod[index2] * oneodetpixscale_lod2);
	    CRbd = CRbdfactor * det_npix; //detector noise
	    CRb += CRbd;
	    twoCRb = 2 * CRb;

	    //*** can delete the following two lines. They're just temporary notes.
	    //junk = CRbbinfactor * skytrans[index] * omega_lod[index2];
	    //CRbbinfactor = Fbinary * oneopixscale2 * FATDL; //count rate for scattered light from nearby stars 

	    //Note: for a 2D array entry c[ix,iy], the 1D index = ix + iy*xsize
	    //Here, ix = iplanet, iy = istar, xsize = np
	    cp = ((CRp + twoCRb) / (CRp*CRp - SNRCRpfloor*SNRCRpfloor)); //this includes the systematic noise floor term a la Bijan Nemati    
	    temptp = SNR2 * cp * toverhead_multi + toverhead_fixed; //record exposure time with overheads


	    //Astrophysical fluxes
	    if(printcountrates == 1 && iorbit == 0 && iphase == 0) printf("%d | Astrophysical:  %e %e %e %e\n",istar,Fstar*F0,Fstar*F0*Fp0[iplanetpistartnp],Fzodi*F0,Fexozodi*F0/(sp[iplanetpistartnp]*sp[iplanetpistartnp]*dist*dist));
	    //Intermediate values
	    if(printcountrates == 1 && iorbit == 0 && iphase == 0) printf("%d | Intermediate:   %e %e %e %e %e %e %e | %e\n",istar,sp_lod,omega_lod[index2],photap_frac[index2],Istar_interp[index]*oneopixscale2,skytrans[index]*oneopixscale2,det_npix,t_photon_count,CRbbinfactor);//* omega_lod[index2]);
	    //Count rates
	    if(printcountrates == 1 && iorbit == 0 && iphase == 0) printf("%d | Count rates:    %e %e %e %e %e %e %e\n",istar,CRp,tempCRbsfactor* omega_lod[index2], tempCRbzfactor * omega_lod[index2], tempCRbezfactor * omega_lod[index2], tempCRbbinfactor * omega_lod[index2], CRbd,SNRCRpfloor/SNR);
	    //Exposure times
	    if(printcountrates == 1 && iorbit == 0 && iphase == 0) printf("%d | Exposure times: %e %e\n",istar,SNR2*cp,temptp);

	    

	    if(temptp < 0) temptp = INF; //It's past the systematic noise floor limit
	    if(temptp > tlimit) temptp = INF; //treat as unobservable if beyond exposure time limit
	    if(nrolls != 1) temptp *= nrolls; //multiply by # of required rolls to achieve 360 deg coverage (after tlimit enforcement)
 
	  } else {
	    //It's below the systematic noise floor...
	    temptp = INF; //treat it as unobservable
	  }
	  
	  if(temptp < besttp_v_ratio) {
	    besttp_v_ratio = temptp; //record the minimum exposure time as a function of PSF truncation ratio
	    bestpsfomega = omega_lod[index2];
	    bestpsftruncratio = psf_trunc_ratio[iratio];
	  }

	  
	} //end of PSF ratio loop
	
	tp[iplanetpistartnp] = besttp_v_ratio; //record the best exposure time
	psfomegas[iplanet] = bestpsfomega; //remember the best omega value
	psftruncratios[iplanet] = bestpsftruncratio; //remember the best truncation value
		
	if(tp[iplanetpistartnp] < mint_v_phase) {
	  mint_v_phase = tp[iplanetpistartnp]; //record the minimum exposure time as a function of phase
	  best_psfomega_v_phase = psfomegas[iplanet]; //record the psf omega associated with the min exposure time as a function of phase
	  best_psftruncratio_v_phase = psftruncratios[iplanet]; //record the psf omega associated with the min exposure time as a function of phase
	}
	
      } else { //outside of the input contrast map or hard IWA/OWA cutoffs
	tp[iplanetpistartnp] = INF; //treat it as unobservable
      }

      
    } //end of phase loop
    if(optimize_phase > 0 && isfinite(mint_v_phase)) { //if user wants the exposure times phase-optimized...
      for(iphase=0;iphase<nphases;iphase++) {
	iplanet = iphase + iorbit*nphases;
	tp[iplanet+istartnp] = mint_v_phase; //set all exposure times to the phase-optimized time
	psfomegas[iplanet] = best_psfomega_v_phase;
	psftruncratios[iplanet] = best_psftruncratio_v_phase;
      }
    }
  } //end of orbits loop
  
  //t5 = clock();

  //Calculate some metrics about PSF omega
  avgpsfomega[istar] = 0.0;
  minpsfomega[istar] = 1e20;
  maxpsfomega[istar] = 0.0;
  avgpsftruncratio[istar] = 0.0;
  minpsftruncratio[istar] = 1e20;
  maxpsftruncratio[istar] = 0.0;
  psfcounter = 0;
  for(iplanet=0;iplanet<np;iplanet++) {
    if(psftruncratios[iplanet] > 1e-5) {
      avgpsfomega[istar] += psfomegas[iplanet];
      avgpsftruncratio[istar] += psftruncratios[iplanet];
      psfcounter++;
      if(psfomegas[iplanet] < minpsfomega[istar]) minpsfomega[istar] = psfomegas[iplanet];
      if(psfomegas[iplanet] > maxpsfomega[istar]) maxpsfomega[istar] = psfomegas[iplanet];      
      if(psftruncratios[iplanet] < minpsftruncratio[istar]) minpsftruncratio[istar] = psftruncratios[iplanet];
      if(psftruncratios[iplanet] > maxpsftruncratio[istar]) maxpsftruncratio[istar] = psftruncratios[iplanet];      
    }
  }
  avgpsfomega[istar] /= psfcounter;
  avgpsftruncratio[istar] /= psfcounter;

  //t6 = clock();

  //Create a temporary copy of the time vector and sort it
  for(iplanet=0;iplanet<np;iplanet++) temp_tp[iplanet] = tp[iplanet+istartnp];      
  qsort(temp_tp, np, sizeof(temp_tp[0]), compare_t); //sort it
  for(iplanet=0;iplanet<np;iplanet++) {
    if(isfinite(temp_tp[iplanet])) { //infinite exp times are now at the end
      cot = ((double) iplanet) / temp_tp[iplanet]; //if it's finite, calculate first visit completeness/time
      if(cot > maxcot) { //if this is the best Completeness/Time value found yet...
      maxcot = cot; //find the peak of first visit Completeness/Time
      }
    } else {
      break;
    }
  }

  //t7 = clock();
  
  free(Istar_interp);
  free(noisefloor_interp);
  free(temp_tp);
  free(psfomegas);
  free(psftruncratios);

  //t8 = clock();

  //if(istar==3) printf("%e %e %e %e %e %e %e %e\n",((double) (t1-t0))/CLOCKS_PER_SEC,((double) (t2-t1))/CLOCKS_PER_SEC,((double) (t3-t2))/CLOCKS_PER_SEC,((double) (t4-t3))/CLOCKS_PER_SEC,((double) (t5-t4))/CLOCKS_PER_SEC,((double) (t6-t5))/CLOCKS_PER_SEC,((double) (t7-t6))/CLOCKS_PER_SEC,((double) (t8-t7))/CLOCKS_PER_SEC);
  
  return maxcot;
}












int exposure_time_calculator(int argc, void *argv[]) {


  //exposure_time_calculator.c
  //Created by Christopher Stark, NASA GSFC
  //Last Updated 5/2/24
  

  //Here are the variables passed by the IDL wrapper.
  //ntarg = # of target stars
  //Lstar = vector (length ntarg) of stellar luminosities (L_solar)
  //dist = vector (length ntarg) of stellar distances (pc)
  //magstar = array (ntarg x nlambda) of apparent stellar magnitudes at wavelength lambda
  //Fzodi_list = array (ntarg x nlambda) of zodi fluxes for each target star
  //Fexozodi_list = array (ntarg x nlambda) of exozodi flux for each target star
  //Fbinary_list = array (ntarg x nlambda) of scattered light from off-axis stellar companions
  //np = # of synthetic exoplanets
  //Fp0 = 3D array (nmeananom, noribts, ntargs)
  //sp = 3D array (nmeananom, norbits, ntargs) of projected on-sky separations (arcsec)
  //xp = 3D array (nmeananom, norbits, ntargs) of x-component of sp (arcsec)
  //yp = 3D array (nmeananom, norbits, ntargs) of y-component of sp (arcsec)
  //SNR = vector (nlambda) desired signal to noise ratio
  //lambda = vector (nlambda) wavelength (microns)
  //deltalambda = vector (nlambda) bandpass (microns)
  //F0 = vector (nlambda) specific flux for zero magnitude at wavelength of lambda (photons cm^-2 nm^-1 s^-1)
  //throughput = vector (nlambda) all throughput terms (reflections, QE, etc.) except for coronagraph mathematical throughput
  //D = telescope diameter (m)
  //A = collecting area (m^2)
  //IWA = smallest possible working angle (l/D)
  //photap_rad = scalar photometric aperture radius (l/D)
  //Istar1, noisefloor1, photap_frac1, skytrans1, pixscale1, npix1, xcenter1, ycenter1: coronagraph inputs
  //....2
  //....3
  //....4
  //tp = vector (length np) of exposure times of each synthetic planet

  
  //%%%%%%%%%%%%%%%%%%%%%%
  // INPUT
  //%%%%%%%%%%%%%%%%%%%%%%

  
  long j = 0;
  //Scalars
  int ns = *(int *) argv[j]; j++; //number of target stars
  int nlambda = *(int *) argv[j]; j++; //number of wavelengths
  
  //Vectors with ns elements
  double *Lstar = (double *) argv[j]; j++; //a vector
  double *dist = (double *) argv[j]; j++; //a vector
  double *angdiamstar_arcsec = (double *) argv[j]; j++; //a vector
  double *min_deltamag = (double *) argv[j]; j++; //a vector

  //Arrays of size ns x nlambda
  double *magstar = (double *) argv[j]; j++; //a vector of length ns x nlambda
  double *Fzodi_list = (double *) argv[j]; j++; //a vector of length ns x nlambda
  double *Fexozodi_list = (double *) argv[j]; j++; //a vector of length ns x nlambda
  double *Fbinary_list = (double *) argv[j]; j++; //a vector of length ns x nlambda

  //Vectors with np elements
  int np = *(int *) argv[j]; j++; //number of synthetic exoplanets
  int nphases = *(int *) argv[j]; j++;
  int norbits = *(int *) argv[j]; j++;
  double *Fp0 = (double *) argv[j]; j++; //a 3D array (nmeananom, norbits, ns)
  double *sp = (double *) argv[j]; j++; //a 3D array (nmeananom, norbits, ns), planet separations in AU
  double *xp = (double *) argv[j]; j++; //x component of s, a 3D array (nmeananom, norbits, ns)
  double *yp = (double *) argv[j]; j++; //y component of s, a 3D array (nmeananom, norbits, ns)

  //Vectors with nlambda elements
  double *lambda = (double *) argv[j]; j++; //a vector of length nlambda
  double *R = (double *) argv[j]; j++; //a vector of length nlambda
  double *SNR = (double *) argv[j]; j++; //a vector of length nlambda
  double *F0 = (double *) argv[j]; j++; //a vector of length nlambda
  double *throughput = (double *) argv[j]; j++; //a vector of length nlambda

  double det_pixscale_mas = *(double *) argv[j]; j++; //a scalar
  double *det_npix_multiplier = (double *) argv[j]; j++; //a vector of length nlambda
  double *det_DC = (double *) argv[j]; j++; //a vector of length nlambda
  double *det_RN = (double *) argv[j]; j++; //a vector of length nlambda
  double *det_tread = (double *) argv[j]; j++; //a vector of length nlambda
  double *det_CIC = (double *) argv[j]; j++; //a vector of length nlambda


  //Scalars
  double D = *(double *) argv[j]; j++; //a scalar
  double A = *(double *) argv[j]; j++; //a scalar
  double IWA = *(double *) argv[j]; j++; //a scalar
  double OWA = *(double *) argv[j]; j++; //a scalar
  double photap_rad = *(double *) argv[j]; j++; //a scalar

  double *psf_trunc_ratio = (double *) argv[j]; j++; //a vector of length npsfratios

  //Coronagraph inputs
  int enable_coro1 = *(int *) argv[j]; j++;  
  double *Istar1 = (double *) argv[j]; j++; //this should be a 3D array for the corongraph (x, y, angdiam)
  double *noisefloor1 = (double *) argv[j]; j++; //this should be a 3D array for the corongraph (x, y, angdiam)
  double *photap_frac1 = (double *) argv[j]; j++; //this should be a 3D array for the coronagraph (x,y,PSF_trunc_ratio)
  double *omega_lod1 = (double *) argv[j]; j++; //this should be a 3D array for the coronagraph (x,y,PSF_trunc_ratio)
  double *skytrans1 = (double *) argv[j]; j++; //this should be a 2D array for the coronagraph (x,y)
  double pixscale1 = *(double *) argv[j]; j++; //size of photap_frac and contrast array pixels in units of lambda/D
  int npix1 = *(int *) argv[j]; j++;
  double xcenter1 = *(double *) argv[j]; j++; //pixel location of the coronagraph center of the Istar and photap_frac arrays
  double ycenter1 = *(double *) argv[j]; j++;
  double bw1 = *(double *) argv[j]; j++;
  double *angdiams1 = (double *) argv[j]; j++; //this should be a vector of length ndiams
  int ndiams1 = *(int *) argv[j]; j++;
  int npsfratios1 = *(int *) argv[j]; j++;
  int nrolls1 = *(int *) argv[j]; j++;
  
  int enable_coro2 = *(int *) argv[j]; j++;  
  double *Istar2 = (double *) argv[j]; j++; //this should be a 3D array for the corongraph (x, y, angdiam)
  double *noisefloor2 = (double *) argv[j]; j++; //this should be a 3D array for the corongraph (x, y, angdiam)
  double *photap_frac2 = (double *) argv[j]; j++; //this should be a 3D array for the coronagraph (x,y,PSF_trunc_ratio)
  double *omega_lod2 = (double *) argv[j]; j++; //this should be a 3D array for the coronagraph (x,y,PSF_trunc_ratio)
  double *skytrans2 = (double *) argv[j]; j++; //this should be a 2D array for the coronagraph (x,y)
  double pixscale2 = *(double *) argv[j]; j++; //size of photap_frac and contrast array pixels in units of lambda/D
  int npix2 = *(int *) argv[j]; j++;
  double xcenter2 = *(double *) argv[j]; j++; //pixel location of the coronagraph center of the Istar and photap_frac arrays
  double ycenter2 = *(double *) argv[j]; j++;
  double bw2 = *(double *) argv[j]; j++;
  double *angdiams2 = (double *) argv[j]; j++; //this should be a vector of length ndiams
  int ndiams2 = *(int *) argv[j]; j++;
  int npsfratios2 = *(int *) argv[j]; j++;
  int nrolls2 = *(int *) argv[j]; j++;

  int enable_coro3 = *(int *) argv[j]; j++;  
  double *Istar3 = (double *) argv[j]; j++; //this should be a 3D array for the corongraph (x, y, angdiam)
  double *noisefloor3 = (double *) argv[j]; j++; //this should be a 3D array for the corongraph (x, y, angdiam)
  double *photap_frac3 = (double *) argv[j]; j++; //this should be a 3D array for the coronagraph (x,y,PSF_trunc_ratio)
  double *omega_lod3 = (double *) argv[j]; j++; //this should be a 3D array for the coronagraph (x,y,PSF_trunc_ratio)
  double *skytrans3 = (double *) argv[j]; j++; //this should be a 2D array for the coronagraph (x,y)
  double pixscale3 = *(double *) argv[j]; j++; //size of photap_frac and contrast array pixels in units of lambda/D
  int npix3 = *(int *) argv[j]; j++;
  double xcenter3 = *(double *) argv[j]; j++; //pixel location of the coronagraph center of the Istar and photap_frac arrays
  double ycenter3 = *(double *) argv[j]; j++;
  double bw3 = *(double *) argv[j]; j++;
  double *angdiams3 = (double *) argv[j]; j++; //this should be a vector of length ndiams
  int ndiams3 = *(int *) argv[j]; j++;
  int npsfratios3 = *(int *) argv[j]; j++;
  int nrolls3 = *(int *) argv[j]; j++;

  int enable_coro4 = *(int *) argv[j]; j++;  
  double *Istar4 = (double *) argv[j]; j++; //this should be a 3D array for the corongraph (x, y, angdiam)
  double *noisefloor4 = (double *) argv[j]; j++; //this should be a 3D array for the corongraph (x, y, angdiam)
  double *photap_frac4 = (double *) argv[j]; j++; //this should be a 3D array for the coronagraph (x,y,PSF_trunc_ratio)
  double *omega_lod4 = (double *) argv[j]; j++; //this should be a 3D array for the coronagraph (x,y,PSF_trunc_ratio)
  double *skytrans4 = (double *) argv[j]; j++; //this should be a 2D array for the coronagraph (x,y)
  double pixscale4 = *(double *) argv[j]; j++; //size of photap_frac and contrast array pixels in units of lambda/D
  int npix4 = *(int *) argv[j]; j++;
  double xcenter4 = *(double *) argv[j]; j++; //pixel location of the coronagraph center of the Istar and photap_frac arrays
  double ycenter4 = *(double *) argv[j]; j++;
  double bw4 = *(double *) argv[j]; j++;
  double *angdiams4 = (double *) argv[j]; j++; //this should be a vector of length ndiams
  int ndiams4 = *(int *) argv[j]; j++;
  int npsfratios4 = *(int *) argv[j]; j++;
  int nrolls4 = *(int *) argv[j]; j++;


  
  double toverhead_fixed = *(double *) argv[j]; j++; //scalar
  double toverhead_multi = *(double *) argv[j]; j++; //scalar
  double tlimit = *(double *) argv[j]; j++; //scalar
  int optimize_phase = *(int *) argv[j]; j++;
  int nooptimize = *(int *) argv[j]; j++;
  int printcountrates = 0;

  //%%%%%%%%%%%%%%%%%%%%%%
  // OUTPUT
  //%%%%%%%%%%%%%%%%%%%%%%

  int *besticoro = (int *) argv[j]; j++; //a vector of length ns
  int *bestilambda = (int *) argv[j]; j++; //a vector of length ns
  double *avgpsfomega = (double *) argv[j]; j++; //a vector of length ns
  double *minpsfomega = (double *) argv[j]; j++; //a vector of length ns
  double *maxpsfomega = (double *) argv[j]; j++; //a vector of length ns
  double *avgpsftruncratio = (double *) argv[j]; j++; //a vector of length ns
  double *minpsftruncratio = (double *) argv[j]; j++; //a vector of length ns
  double *maxpsftruncratio = (double *) argv[j]; j++; //a vector of length ns
  double *tp = (double *) argv[j]; j++; //a 3D array [nmeananom, norbits, ns]


  //%%%%%%%%%%%%%%%%%%%%%%
  // LOCAL VARIABLES
  //%%%%%%%%%%%%%%%%%%%%%%


  int i, istar, icoro, ilambda;
  int result, run_this_coro;
  double *Istar, *noisefloor, *photap_frac, *omega_lod, *skytrans, *angdiams;
  double xcenter, ycenter, bw, pixscale;
  int npix, ndiams, npsfratios, nrolls;
  double cot, maxcot;


  if(nlambda > 20) {
    printf("nlambda must be <= MAXNLAMBDA. Quitting...\n");
    exit(1);
  }
  

  //%%%%%%%%%%%%%%%%%%%%%%
  // MAIN LOOP
  //%%%%%%%%%%%%%%%%%%%%%%
  
  
  //Loop through each target star
#pragma omp parallel private(result, Istar, noisefloor, photap_frac, omega_lod, skytrans, pixscale, npix, xcenter, ycenter, bw, angdiams, ndiams, npsfratios, nrolls, run_this_coro, icoro, ilambda, cot, maxcot, printcountrates)
  { 


    
    #pragma omp for
    for(istar=0;istar<ns;istar++) {

      cot = 0.0;
      maxcot = 0.0;

      for(icoro=0;icoro<4;icoro++) {

	
	if(nooptimize > 0) icoro=besticoro[istar]; //if no optimization is desired, set icoro to the input value
	
	//Set the pointers to the right coronagraph model
	run_this_coro = 0;
	if(icoro==0 && enable_coro1==1) {
	  run_this_coro = 1;
	  Istar=Istar1; noisefloor=noisefloor1; photap_frac=photap_frac1; omega_lod=omega_lod1; skytrans=skytrans1; pixscale=pixscale1; 
	  npix=npix1; xcenter=xcenter1; ycenter=ycenter1; bw=bw1; angdiams=angdiams1; ndiams=ndiams1; npsfratios=npsfratios1; nrolls=nrolls1;
	}
	if(icoro==1 && enable_coro2==1) {
	  run_this_coro = 1;
	  Istar=Istar2; noisefloor=noisefloor2; photap_frac=photap_frac2; omega_lod=omega_lod2; skytrans=skytrans2; pixscale=pixscale2; 
	  npix=npix2; xcenter=xcenter2; ycenter=ycenter2; bw=bw2; angdiams=angdiams2; ndiams=ndiams2; npsfratios=npsfratios2; nrolls=nrolls2;
	}
	if(icoro==2 && enable_coro3==1) {
	  run_this_coro = 1;
	  Istar=Istar3; noisefloor=noisefloor3; photap_frac=photap_frac3; omega_lod=omega_lod3; skytrans=skytrans3; pixscale=pixscale3; 
	  npix=npix3; xcenter=xcenter3; ycenter=ycenter3; bw=bw3; angdiams=angdiams3; ndiams=ndiams3; npsfratios=npsfratios3; nrolls=nrolls3;
	}
	if(icoro==3 && enable_coro4==1) {
	  run_this_coro = 1;
	  Istar=Istar4; noisefloor=noisefloor4; photap_frac=photap_frac4; omega_lod=omega_lod4; skytrans=skytrans4; pixscale=pixscale4; 
	  npix=npix4; xcenter=xcenter4; ycenter=ycenter4; bw=bw4; angdiams=angdiams4; ndiams=ndiams4; npsfratios=npsfratios4; nrolls=nrolls4;
	}


	
	if(run_this_coro == 1) {


	  for(ilambda=0;ilambda<nlambda;ilambda++) {

	    if(nooptimize > 0) ilambda=bestilambda[istar]; //if no optimization is desired, set ilambda to the input value
	    printcountrates=0; //^^^^^^^^^^^^^^
	    
	    cot = calc_exp_time(ns, np, istar, ilambda, nphases, norbits, IWA, OWA, min_deltamag[istar], \
				Istar, noisefloor, photap_frac, omega_lod, skytrans, npix, xcenter, ycenter, pixscale, bw, angdiams, ndiams, npsfratios, nrolls, psf_trunc_ratio,  \
				lambda[ilambda], D, A, magstar[istar + ilambda*ns], F0[ilambda], throughput[ilambda],  \
				Fzodi_list[istar + ilambda*ns], Fexozodi_list[istar + ilambda*ns], Fbinary_list[istar + ilambda*ns], R[ilambda], SNR[ilambda], \
				det_pixscale_mas, det_npix_multiplier[ilambda], det_DC[ilambda], det_RN[ilambda], det_tread[ilambda], det_CIC[ilambda], \
				dist[istar], angdiamstar_arcsec[istar], toverhead_fixed, toverhead_multi, tlimit, optimize_phase, Fp0, sp, xp, yp, tp, \
				avgpsfomega, minpsfomega, maxpsfomega, avgpsftruncratio, minpsftruncratio, maxpsftruncratio, printcountrates);
	    
	    if(cot > maxcot) {
	      maxcot = cot;
	      bestilambda[istar] = ilambda;
	      besticoro[istar] = icoro;
	    }

	    if(nooptimize > 0) ilambda = nlambda; //if no optimization is desired, break the loop

	  } //end of wavelength loop


	}

	if(nooptimize > 0) icoro = 4; //if no optimization is desired, break the loop
	
      } //end of coronagraph loop
      

      if(nooptimize == 0) { //if optimization occurred, we have to do the following

	printcountrates=0; //^^^^^^^^^^^^^^
	
	//Now that we have determined the best coronagraph and wavelength,
	//call the exposure time calculator again using this combination.
	//This makes sure the tp values are up to date.
	ilambda = bestilambda[istar];
	if(besticoro[istar]==0) {
	  if(enable_coro1 != 1) {printf(stderr,"ERROR: coro1 is not enabled.\n"); exit (1);}
	  Istar=Istar1; noisefloor=noisefloor1; photap_frac=photap_frac1; omega_lod=omega_lod1; skytrans=skytrans1; pixscale=pixscale1;
	  npix=npix1; xcenter=xcenter1; ycenter=ycenter1; bw=bw1; angdiams=angdiams1; ndiams=ndiams1; npsfratios=npsfratios1; nrolls=nrolls1;
	}
	if(besticoro[istar]==1) {
	  if(enable_coro2 != 1) {printf(stderr,"ERROR: coro2 is not enabled.\n"); exit (1);}
	  Istar=Istar2; noisefloor=noisefloor2; photap_frac=photap_frac2; omega_lod=omega_lod2; skytrans=skytrans2; pixscale=pixscale2;
	  npix=npix2; xcenter=xcenter2; ycenter=ycenter2; bw=bw2; angdiams=angdiams2; ndiams=ndiams2; npsfratios=npsfratios2; nrolls=nrolls2;
	}
	if(besticoro[istar]==2) {
	  if(enable_coro3 != 1) {printf(stderr,"ERROR: coro3 is not enabled.\n"); exit (1);}
	  Istar=Istar3; noisefloor=noisefloor3; photap_frac=photap_frac3; omega_lod=omega_lod3; skytrans=skytrans3; pixscale=pixscale3;
	  npix=npix3; xcenter=xcenter3; ycenter=ycenter3; bw=bw3; angdiams=angdiams3; ndiams=ndiams3; npsfratios=npsfratios3; nrolls=nrolls3;
	}
	if(besticoro[istar]==3) {
	  if(enable_coro4 != 1) {printf(stderr,"ERROR: coro4 is not enabled.\n"); exit (1);}
	  Istar=Istar4; noisefloor=noisefloor4; photap_frac=photap_frac4; omega_lod=omega_lod4; skytrans=skytrans4; pixscale=pixscale4;
	  npix=npix4; xcenter=xcenter4; ycenter=ycenter4; bw=bw4; angdiams=angdiams4; ndiams=ndiams4; npsfratios=npsfratios4; nrolls=nrolls4;
	}

	cot = calc_exp_time(ns, np, istar, ilambda, nphases, norbits, IWA, OWA, min_deltamag[istar], \
			    Istar, noisefloor, photap_frac, omega_lod, skytrans, npix, xcenter, ycenter, pixscale, bw, angdiams, ndiams, npsfratios, nrolls, psf_trunc_ratio, \
			    lambda[ilambda], D, A, magstar[istar + ilambda*ns], F0[ilambda], throughput[ilambda],  \
			    Fzodi_list[istar + ilambda*ns], Fexozodi_list[istar + ilambda*ns], Fbinary_list[istar + ilambda*ns], R[ilambda], SNR[ilambda], \
			    det_pixscale_mas, det_npix_multiplier[ilambda], det_DC[ilambda], det_RN[ilambda], det_tread[ilambda], det_CIC[ilambda], \
			    dist[istar], angdiamstar_arcsec[istar], toverhead_fixed, toverhead_multi, tlimit, optimize_phase, Fp0, sp, xp, yp, tp, \
			    avgpsfomega, minpsfomega, maxpsfomega, avgpsftruncratio, minpsftruncratio, maxpsftruncratio, printcountrates);
	}


      //printf("exp time (h) = %e\n",tp[0]);
    } //end of istar loop
    
    
  } //end of parallelization
  
  return 0;
}





